# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: TextClassifier.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import DataStructures_pb2 as DataStructures__pb2
try:
  FeatureTypes__pb2 = DataStructures__pb2.FeatureTypes__pb2
except AttributeError:
  FeatureTypes__pb2 = DataStructures__pb2.FeatureTypes_pb2

from .DataStructures_pb2 import *

DESCRIPTOR = _descriptor.FileDescriptor(
  name='TextClassifier.proto',
  package='CoreML.Specification.CoreMLModels',
  syntax='proto3',
  serialized_pb=_b('\n\x14TextClassifier.proto\x12!CoreML.Specification.CoreMLModels\x1a\x14\x44\x61taStructures.proto\"\xa1\x01\n\x0eTextClassifier\x12\x10\n\x08revision\x18\x01 \x01(\r\x12\x10\n\x08language\x18\n \x01(\t\x12\x1a\n\x12modelParameterData\x18\x64 \x01(\x0c\x12@\n\x11stringClassLabels\x18\xc8\x01 \x01(\x0b\x32\".CoreML.Specification.StringVectorH\x00\x42\r\n\x0b\x43lassLabelsB\x02H\x03P\x00\x62\x06proto3')
  ,
  dependencies=[DataStructures__pb2.DESCRIPTOR,],
  public_dependencies=[DataStructures__pb2.DESCRIPTOR,])




_TEXTCLASSIFIER = _descriptor.Descriptor(
  name='TextClassifier',
  full_name='CoreML.Specification.CoreMLModels.TextClassifier',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='revision', full_name='CoreML.Specification.CoreMLModels.TextClassifier.revision', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='language', full_name='CoreML.Specification.CoreMLModels.TextClassifier.language', index=1,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='modelParameterData', full_name='CoreML.Specification.CoreMLModels.TextClassifier.modelParameterData', index=2,
      number=100, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='stringClassLabels', full_name='CoreML.Specification.CoreMLModels.TextClassifier.stringClassLabels', index=3,
      number=200, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='ClassLabels', full_name='CoreML.Specification.CoreMLModels.TextClassifier.ClassLabels',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=82,
  serialized_end=243,
)

_TEXTCLASSIFIER.fields_by_name['stringClassLabels'].message_type = DataStructures__pb2._STRINGVECTOR
_TEXTCLASSIFIER.oneofs_by_name['ClassLabels'].fields.append(
  _TEXTCLASSIFIER.fields_by_name['stringClassLabels'])
_TEXTCLASSIFIER.fields_by_name['stringClassLabels'].containing_oneof = _TEXTCLASSIFIER.oneofs_by_name['ClassLabels']
DESCRIPTOR.message_types_by_name['TextClassifier'] = _TEXTCLASSIFIER
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

TextClassifier = _reflection.GeneratedProtocolMessageType('TextClassifier', (_message.Message,), dict(
  DESCRIPTOR = _TEXTCLASSIFIER,
  __module__ = 'TextClassifier_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.CoreMLModels.TextClassifier)
  ))
_sym_db.RegisterMessage(TextClassifier)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(descriptor_pb2.FileOptions(), _b('H\003'))
# @@protoc_insertion_point(module_scope)
