# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: SoundAnalysisPreprocessing.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='SoundAnalysisPreprocessing.proto',
  package='CoreML.Specification.CoreMLModels',
  syntax='proto3',
  serialized_pb=_b('\n SoundAnalysisPreprocessing.proto\x12!CoreML.Specification.CoreMLModels\"\xa0\x01\n\x1aSoundAnalysisPreprocessing\x12V\n\x06vggish\x18\x14 \x01(\x0b\x32\x44.CoreML.Specification.CoreMLModels.SoundAnalysisPreprocessing.VggishH\x00\x1a\x08\n\x06VggishB \n\x1eSoundAnalysisPreprocessingTypeB\x02H\x03\x62\x06proto3')
)




_SOUNDANALYSISPREPROCESSING_VGGISH = _descriptor.Descriptor(
  name='Vggish',
  full_name='CoreML.Specification.CoreMLModels.SoundAnalysisPreprocessing.Vggish',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=190,
  serialized_end=198,
)

_SOUNDANALYSISPREPROCESSING = _descriptor.Descriptor(
  name='SoundAnalysisPreprocessing',
  full_name='CoreML.Specification.CoreMLModels.SoundAnalysisPreprocessing',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='vggish', full_name='CoreML.Specification.CoreMLModels.SoundAnalysisPreprocessing.vggish', index=0,
      number=20, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_SOUNDANALYSISPREPROCESSING_VGGISH, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='SoundAnalysisPreprocessingType', full_name='CoreML.Specification.CoreMLModels.SoundAnalysisPreprocessing.SoundAnalysisPreprocessingType',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=72,
  serialized_end=232,
)

_SOUNDANALYSISPREPROCESSING_VGGISH.containing_type = _SOUNDANALYSISPREPROCESSING
_SOUNDANALYSISPREPROCESSING.fields_by_name['vggish'].message_type = _SOUNDANALYSISPREPROCESSING_VGGISH
_SOUNDANALYSISPREPROCESSING.oneofs_by_name['SoundAnalysisPreprocessingType'].fields.append(
  _SOUNDANALYSISPREPROCESSING.fields_by_name['vggish'])
_SOUNDANALYSISPREPROCESSING.fields_by_name['vggish'].containing_oneof = _SOUNDANALYSISPREPROCESSING.oneofs_by_name['SoundAnalysisPreprocessingType']
DESCRIPTOR.message_types_by_name['SoundAnalysisPreprocessing'] = _SOUNDANALYSISPREPROCESSING
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

SoundAnalysisPreprocessing = _reflection.GeneratedProtocolMessageType('SoundAnalysisPreprocessing', (_message.Message,), dict(

  Vggish = _reflection.GeneratedProtocolMessageType('Vggish', (_message.Message,), dict(
    DESCRIPTOR = _SOUNDANALYSISPREPROCESSING_VGGISH,
    __module__ = 'SoundAnalysisPreprocessing_pb2'
    # @@protoc_insertion_point(class_scope:CoreML.Specification.CoreMLModels.SoundAnalysisPreprocessing.Vggish)
    ))
  ,
  DESCRIPTOR = _SOUNDANALYSISPREPROCESSING,
  __module__ = 'SoundAnalysisPreprocessing_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.CoreMLModels.SoundAnalysisPreprocessing)
  ))
_sym_db.RegisterMessage(SoundAnalysisPreprocessing)
_sym_db.RegisterMessage(SoundAnalysisPreprocessing.Vggish)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(descriptor_pb2.FileOptions(), _b('H\003'))
# @@protoc_insertion_point(module_scope)
