# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: GLMRegressor.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='GLMRegressor.proto',
  package='CoreML.Specification',
  syntax='proto3',
  serialized_pb=_b('\n\x12GLMRegressor.proto\x12\x14\x43oreML.Specification\"\x9d\x02\n\x0cGLMRegressor\x12?\n\x07weights\x18\x01 \x03(\x0b\x32..CoreML.Specification.GLMRegressor.DoubleArray\x12\x0e\n\x06offset\x18\x02 \x03(\x01\x12[\n\x17postEvaluationTransform\x18\x03 \x01(\x0e\x32:.CoreML.Specification.GLMRegressor.PostEvaluationTransform\x1a\x1c\n\x0b\x44oubleArray\x12\r\n\x05value\x18\x01 \x03(\x01\"A\n\x17PostEvaluationTransform\x12\x0f\n\x0bNoTransform\x10\x00\x12\t\n\x05Logit\x10\x01\x12\n\n\x06Probit\x10\x02\x42\x02H\x03\x62\x06proto3')
)



_GLMREGRESSOR_POSTEVALUATIONTRANSFORM = _descriptor.EnumDescriptor(
  name='PostEvaluationTransform',
  full_name='CoreML.Specification.GLMRegressor.PostEvaluationTransform',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='NoTransform', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='Logit', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='Probit', index=2, number=2,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=265,
  serialized_end=330,
)
_sym_db.RegisterEnumDescriptor(_GLMREGRESSOR_POSTEVALUATIONTRANSFORM)


_GLMREGRESSOR_DOUBLEARRAY = _descriptor.Descriptor(
  name='DoubleArray',
  full_name='CoreML.Specification.GLMRegressor.DoubleArray',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='value', full_name='CoreML.Specification.GLMRegressor.DoubleArray.value', index=0,
      number=1, type=1, cpp_type=5, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=235,
  serialized_end=263,
)

_GLMREGRESSOR = _descriptor.Descriptor(
  name='GLMRegressor',
  full_name='CoreML.Specification.GLMRegressor',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='weights', full_name='CoreML.Specification.GLMRegressor.weights', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='offset', full_name='CoreML.Specification.GLMRegressor.offset', index=1,
      number=2, type=1, cpp_type=5, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='postEvaluationTransform', full_name='CoreML.Specification.GLMRegressor.postEvaluationTransform', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_GLMREGRESSOR_DOUBLEARRAY, ],
  enum_types=[
    _GLMREGRESSOR_POSTEVALUATIONTRANSFORM,
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=45,
  serialized_end=330,
)

_GLMREGRESSOR_DOUBLEARRAY.containing_type = _GLMREGRESSOR
_GLMREGRESSOR.fields_by_name['weights'].message_type = _GLMREGRESSOR_DOUBLEARRAY
_GLMREGRESSOR.fields_by_name['postEvaluationTransform'].enum_type = _GLMREGRESSOR_POSTEVALUATIONTRANSFORM
_GLMREGRESSOR_POSTEVALUATIONTRANSFORM.containing_type = _GLMREGRESSOR
DESCRIPTOR.message_types_by_name['GLMRegressor'] = _GLMREGRESSOR
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

GLMRegressor = _reflection.GeneratedProtocolMessageType('GLMRegressor', (_message.Message,), dict(

  DoubleArray = _reflection.GeneratedProtocolMessageType('DoubleArray', (_message.Message,), dict(
    DESCRIPTOR = _GLMREGRESSOR_DOUBLEARRAY,
    __module__ = 'GLMRegressor_pb2'
    # @@protoc_insertion_point(class_scope:CoreML.Specification.GLMRegressor.DoubleArray)
    ))
  ,
  DESCRIPTOR = _GLMREGRESSOR,
  __module__ = 'GLMRegressor_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.GLMRegressor)
  ))
_sym_db.RegisterMessage(GLMRegressor)
_sym_db.RegisterMessage(GLMRegressor.DoubleArray)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(descriptor_pb2.FileOptions(), _b('H\003'))
# @@protoc_insertion_point(module_scope)
