# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: CategoricalMapping.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import DataStructures_pb2 as DataStructures__pb2
try:
  FeatureTypes__pb2 = DataStructures__pb2.FeatureTypes__pb2
except AttributeError:
  FeatureTypes__pb2 = DataStructures__pb2.FeatureTypes_pb2

from .DataStructures_pb2 import *

DESCRIPTOR = _descriptor.FileDescriptor(
  name='CategoricalMapping.proto',
  package='CoreML.Specification',
  syntax='proto3',
  serialized_pb=_b('\n\x18\x43\x61tegoricalMapping.proto\x12\x14\x43oreML.Specification\x1a\x14\x44\x61taStructures.proto\"\xe7\x01\n\x12\x43\x61tegoricalMapping\x12\x42\n\x10stringToInt64Map\x18\x01 \x01(\x0b\x32&.CoreML.Specification.StringToInt64MapH\x00\x12\x42\n\x10int64ToStringMap\x18\x02 \x01(\x0b\x32&.CoreML.Specification.Int64ToStringMapH\x00\x12\x12\n\x08strValue\x18\x65 \x01(\tH\x01\x12\x14\n\nint64Value\x18\x66 \x01(\x03H\x01\x42\r\n\x0bMappingTypeB\x10\n\x0eValueOnUnknownB\x02H\x03P\x00\x62\x06proto3')
  ,
  dependencies=[DataStructures__pb2.DESCRIPTOR,],
  public_dependencies=[DataStructures__pb2.DESCRIPTOR,])




_CATEGORICALMAPPING = _descriptor.Descriptor(
  name='CategoricalMapping',
  full_name='CoreML.Specification.CategoricalMapping',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='stringToInt64Map', full_name='CoreML.Specification.CategoricalMapping.stringToInt64Map', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='int64ToStringMap', full_name='CoreML.Specification.CategoricalMapping.int64ToStringMap', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='strValue', full_name='CoreML.Specification.CategoricalMapping.strValue', index=2,
      number=101, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='int64Value', full_name='CoreML.Specification.CategoricalMapping.int64Value', index=3,
      number=102, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='MappingType', full_name='CoreML.Specification.CategoricalMapping.MappingType',
      index=0, containing_type=None, fields=[]),
    _descriptor.OneofDescriptor(
      name='ValueOnUnknown', full_name='CoreML.Specification.CategoricalMapping.ValueOnUnknown',
      index=1, containing_type=None, fields=[]),
  ],
  serialized_start=73,
  serialized_end=304,
)

_CATEGORICALMAPPING.fields_by_name['stringToInt64Map'].message_type = DataStructures__pb2._STRINGTOINT64MAP
_CATEGORICALMAPPING.fields_by_name['int64ToStringMap'].message_type = DataStructures__pb2._INT64TOSTRINGMAP
_CATEGORICALMAPPING.oneofs_by_name['MappingType'].fields.append(
  _CATEGORICALMAPPING.fields_by_name['stringToInt64Map'])
_CATEGORICALMAPPING.fields_by_name['stringToInt64Map'].containing_oneof = _CATEGORICALMAPPING.oneofs_by_name['MappingType']
_CATEGORICALMAPPING.oneofs_by_name['MappingType'].fields.append(
  _CATEGORICALMAPPING.fields_by_name['int64ToStringMap'])
_CATEGORICALMAPPING.fields_by_name['int64ToStringMap'].containing_oneof = _CATEGORICALMAPPING.oneofs_by_name['MappingType']
_CATEGORICALMAPPING.oneofs_by_name['ValueOnUnknown'].fields.append(
  _CATEGORICALMAPPING.fields_by_name['strValue'])
_CATEGORICALMAPPING.fields_by_name['strValue'].containing_oneof = _CATEGORICALMAPPING.oneofs_by_name['ValueOnUnknown']
_CATEGORICALMAPPING.oneofs_by_name['ValueOnUnknown'].fields.append(
  _CATEGORICALMAPPING.fields_by_name['int64Value'])
_CATEGORICALMAPPING.fields_by_name['int64Value'].containing_oneof = _CATEGORICALMAPPING.oneofs_by_name['ValueOnUnknown']
DESCRIPTOR.message_types_by_name['CategoricalMapping'] = _CATEGORICALMAPPING
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

CategoricalMapping = _reflection.GeneratedProtocolMessageType('CategoricalMapping', (_message.Message,), dict(
  DESCRIPTOR = _CATEGORICALMAPPING,
  __module__ = 'CategoricalMapping_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.CategoricalMapping)
  ))
_sym_db.RegisterMessage(CategoricalMapping)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(descriptor_pb2.FileOptions(), _b('H\003'))
# @@protoc_insertion_point(module_scope)
