# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: CustomModel.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='CustomModel.proto',
  package='CoreML.Specification',
  syntax='proto3',
  serialized_pb=_b('\n\x11\x43ustomModel.proto\x12\x14\x43oreML.Specification\"\x8d\x03\n\x0b\x43ustomModel\x12\x11\n\tclassName\x18\n \x01(\t\x12\x45\n\nparameters\x18\x1e \x03(\x0b\x32\x31.CoreML.Specification.CustomModel.ParametersEntry\x12\x13\n\x0b\x64\x65scription\x18( \x01(\t\x1a\xa2\x01\n\x15\x43ustomModelParamValue\x12\x15\n\x0b\x64oubleValue\x18\n \x01(\x01H\x00\x12\x15\n\x0bstringValue\x18\x14 \x01(\tH\x00\x12\x12\n\x08intValue\x18\x1e \x01(\x05H\x00\x12\x13\n\tlongValue\x18( \x01(\x03H\x00\x12\x13\n\tboolValue\x18\x32 \x01(\x08H\x00\x12\x14\n\nbytesValue\x18< \x01(\x0cH\x00\x42\x07\n\x05value\x1aj\n\x0fParametersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x46\n\x05value\x18\x02 \x01(\x0b\x32\x37.CoreML.Specification.CustomModel.CustomModelParamValue:\x02\x38\x01\x42\x02H\x03\x62\x06proto3')
)




_CUSTOMMODEL_CUSTOMMODELPARAMVALUE = _descriptor.Descriptor(
  name='CustomModelParamValue',
  full_name='CoreML.Specification.CustomModel.CustomModelParamValue',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='doubleValue', full_name='CoreML.Specification.CustomModel.CustomModelParamValue.doubleValue', index=0,
      number=10, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='stringValue', full_name='CoreML.Specification.CustomModel.CustomModelParamValue.stringValue', index=1,
      number=20, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='intValue', full_name='CoreML.Specification.CustomModel.CustomModelParamValue.intValue', index=2,
      number=30, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='longValue', full_name='CoreML.Specification.CustomModel.CustomModelParamValue.longValue', index=3,
      number=40, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='boolValue', full_name='CoreML.Specification.CustomModel.CustomModelParamValue.boolValue', index=4,
      number=50, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='bytesValue', full_name='CoreML.Specification.CustomModel.CustomModelParamValue.bytesValue', index=5,
      number=60, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='value', full_name='CoreML.Specification.CustomModel.CustomModelParamValue.value',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=171,
  serialized_end=333,
)

_CUSTOMMODEL_PARAMETERSENTRY = _descriptor.Descriptor(
  name='ParametersEntry',
  full_name='CoreML.Specification.CustomModel.ParametersEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='CoreML.Specification.CustomModel.ParametersEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='value', full_name='CoreML.Specification.CustomModel.ParametersEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=_descriptor._ParseOptions(descriptor_pb2.MessageOptions(), _b('8\001')),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=335,
  serialized_end=441,
)

_CUSTOMMODEL = _descriptor.Descriptor(
  name='CustomModel',
  full_name='CoreML.Specification.CustomModel',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='className', full_name='CoreML.Specification.CustomModel.className', index=0,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='parameters', full_name='CoreML.Specification.CustomModel.parameters', index=1,
      number=30, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='description', full_name='CoreML.Specification.CustomModel.description', index=2,
      number=40, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_CUSTOMMODEL_CUSTOMMODELPARAMVALUE, _CUSTOMMODEL_PARAMETERSENTRY, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=44,
  serialized_end=441,
)

_CUSTOMMODEL_CUSTOMMODELPARAMVALUE.containing_type = _CUSTOMMODEL
_CUSTOMMODEL_CUSTOMMODELPARAMVALUE.oneofs_by_name['value'].fields.append(
  _CUSTOMMODEL_CUSTOMMODELPARAMVALUE.fields_by_name['doubleValue'])
_CUSTOMMODEL_CUSTOMMODELPARAMVALUE.fields_by_name['doubleValue'].containing_oneof = _CUSTOMMODEL_CUSTOMMODELPARAMVALUE.oneofs_by_name['value']
_CUSTOMMODEL_CUSTOMMODELPARAMVALUE.oneofs_by_name['value'].fields.append(
  _CUSTOMMODEL_CUSTOMMODELPARAMVALUE.fields_by_name['stringValue'])
_CUSTOMMODEL_CUSTOMMODELPARAMVALUE.fields_by_name['stringValue'].containing_oneof = _CUSTOMMODEL_CUSTOMMODELPARAMVALUE.oneofs_by_name['value']
_CUSTOMMODEL_CUSTOMMODELPARAMVALUE.oneofs_by_name['value'].fields.append(
  _CUSTOMMODEL_CUSTOMMODELPARAMVALUE.fields_by_name['intValue'])
_CUSTOMMODEL_CUSTOMMODELPARAMVALUE.fields_by_name['intValue'].containing_oneof = _CUSTOMMODEL_CUSTOMMODELPARAMVALUE.oneofs_by_name['value']
_CUSTOMMODEL_CUSTOMMODELPARAMVALUE.oneofs_by_name['value'].fields.append(
  _CUSTOMMODEL_CUSTOMMODELPARAMVALUE.fields_by_name['longValue'])
_CUSTOMMODEL_CUSTOMMODELPARAMVALUE.fields_by_name['longValue'].containing_oneof = _CUSTOMMODEL_CUSTOMMODELPARAMVALUE.oneofs_by_name['value']
_CUSTOMMODEL_CUSTOMMODELPARAMVALUE.oneofs_by_name['value'].fields.append(
  _CUSTOMMODEL_CUSTOMMODELPARAMVALUE.fields_by_name['boolValue'])
_CUSTOMMODEL_CUSTOMMODELPARAMVALUE.fields_by_name['boolValue'].containing_oneof = _CUSTOMMODEL_CUSTOMMODELPARAMVALUE.oneofs_by_name['value']
_CUSTOMMODEL_CUSTOMMODELPARAMVALUE.oneofs_by_name['value'].fields.append(
  _CUSTOMMODEL_CUSTOMMODELPARAMVALUE.fields_by_name['bytesValue'])
_CUSTOMMODEL_CUSTOMMODELPARAMVALUE.fields_by_name['bytesValue'].containing_oneof = _CUSTOMMODEL_CUSTOMMODELPARAMVALUE.oneofs_by_name['value']
_CUSTOMMODEL_PARAMETERSENTRY.fields_by_name['value'].message_type = _CUSTOMMODEL_CUSTOMMODELPARAMVALUE
_CUSTOMMODEL_PARAMETERSENTRY.containing_type = _CUSTOMMODEL
_CUSTOMMODEL.fields_by_name['parameters'].message_type = _CUSTOMMODEL_PARAMETERSENTRY
DESCRIPTOR.message_types_by_name['CustomModel'] = _CUSTOMMODEL
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

CustomModel = _reflection.GeneratedProtocolMessageType('CustomModel', (_message.Message,), dict(

  CustomModelParamValue = _reflection.GeneratedProtocolMessageType('CustomModelParamValue', (_message.Message,), dict(
    DESCRIPTOR = _CUSTOMMODEL_CUSTOMMODELPARAMVALUE,
    __module__ = 'CustomModel_pb2'
    # @@protoc_insertion_point(class_scope:CoreML.Specification.CustomModel.CustomModelParamValue)
    ))
  ,

  ParametersEntry = _reflection.GeneratedProtocolMessageType('ParametersEntry', (_message.Message,), dict(
    DESCRIPTOR = _CUSTOMMODEL_PARAMETERSENTRY,
    __module__ = 'CustomModel_pb2'
    # @@protoc_insertion_point(class_scope:CoreML.Specification.CustomModel.ParametersEntry)
    ))
  ,
  DESCRIPTOR = _CUSTOMMODEL,
  __module__ = 'CustomModel_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.CustomModel)
  ))
_sym_db.RegisterMessage(CustomModel)
_sym_db.RegisterMessage(CustomModel.CustomModelParamValue)
_sym_db.RegisterMessage(CustomModel.ParametersEntry)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(descriptor_pb2.FileOptions(), _b('H\003'))
_CUSTOMMODEL_PARAMETERSENTRY.has_options = True
_CUSTOMMODEL_PARAMETERSENTRY._options = _descriptor._ParseOptions(descriptor_pb2.MessageOptions(), _b('8\001'))
# @@protoc_insertion_point(module_scope)
