# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: BayesianProbitRegressor.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='BayesianProbitRegressor.proto',
  package='CoreML.Specification',
  syntax='proto3',
  serialized_pb=_b('\n\x1d\x42\x61yesianProbitRegressor.proto\x12\x14\x43oreML.Specification\"\xa0\x06\n\x17\x42\x61yesianProbitRegressor\x12\x18\n\x10numberOfFeatures\x18\x01 \x01(\r\x12\x44\n\x04\x62ias\x18\x02 \x01(\x0b\x32\x36.CoreML.Specification.BayesianProbitRegressor.Gaussian\x12M\n\x08\x66\x65\x61tures\x18\x03 \x03(\x0b\x32;.CoreML.Specification.BayesianProbitRegressor.FeatureWeight\x12\"\n\x1aregressionInputFeatureName\x18\n \x01(\t\x12 \n\x18optimismInputFeatureName\x18\x0b \x01(\t\x12%\n\x1dsamplingScaleInputFeatureName\x18\x0c \x01(\t\x12*\n\"samplingTruncationInputFeatureName\x18\r \x01(\t\x12\x1d\n\x15meanOutputFeatureName\x18\x14 \x01(\t\x12!\n\x19varianceOutputFeatureName\x18\x15 \x01(\t\x12/\n\'pessimisticProbabilityOutputFeatureName\x18\x16 \x01(\t\x12+\n#sampledProbabilityOutputFeatureName\x18\x17 \x01(\t\x1a+\n\x08Gaussian\x12\x0c\n\x04mean\x18\x01 \x01(\x01\x12\x11\n\tprecision\x18\x02 \x01(\x01\x1ay\n\x12\x46\x65\x61tureValueWeight\x12\x14\n\x0c\x66\x65\x61tureValue\x18\x01 \x01(\r\x12M\n\rfeatureWeight\x18\x02 \x01(\x0b\x32\x36.CoreML.Specification.BayesianProbitRegressor.Gaussian\x1au\n\rFeatureWeight\x12\x11\n\tfeatureId\x18\x01 \x01(\r\x12Q\n\x07weights\x18\x02 \x03(\x0b\x32@.CoreML.Specification.BayesianProbitRegressor.FeatureValueWeightB\x02H\x03\x62\x06proto3')
)




_BAYESIANPROBITREGRESSOR_GAUSSIAN = _descriptor.Descriptor(
  name='Gaussian',
  full_name='CoreML.Specification.BayesianProbitRegressor.Gaussian',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='mean', full_name='CoreML.Specification.BayesianProbitRegressor.Gaussian.mean', index=0,
      number=1, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='precision', full_name='CoreML.Specification.BayesianProbitRegressor.Gaussian.precision', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=571,
  serialized_end=614,
)

_BAYESIANPROBITREGRESSOR_FEATUREVALUEWEIGHT = _descriptor.Descriptor(
  name='FeatureValueWeight',
  full_name='CoreML.Specification.BayesianProbitRegressor.FeatureValueWeight',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='featureValue', full_name='CoreML.Specification.BayesianProbitRegressor.FeatureValueWeight.featureValue', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='featureWeight', full_name='CoreML.Specification.BayesianProbitRegressor.FeatureValueWeight.featureWeight', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=616,
  serialized_end=737,
)

_BAYESIANPROBITREGRESSOR_FEATUREWEIGHT = _descriptor.Descriptor(
  name='FeatureWeight',
  full_name='CoreML.Specification.BayesianProbitRegressor.FeatureWeight',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='featureId', full_name='CoreML.Specification.BayesianProbitRegressor.FeatureWeight.featureId', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='weights', full_name='CoreML.Specification.BayesianProbitRegressor.FeatureWeight.weights', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=739,
  serialized_end=856,
)

_BAYESIANPROBITREGRESSOR = _descriptor.Descriptor(
  name='BayesianProbitRegressor',
  full_name='CoreML.Specification.BayesianProbitRegressor',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='numberOfFeatures', full_name='CoreML.Specification.BayesianProbitRegressor.numberOfFeatures', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='bias', full_name='CoreML.Specification.BayesianProbitRegressor.bias', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='features', full_name='CoreML.Specification.BayesianProbitRegressor.features', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='regressionInputFeatureName', full_name='CoreML.Specification.BayesianProbitRegressor.regressionInputFeatureName', index=3,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='optimismInputFeatureName', full_name='CoreML.Specification.BayesianProbitRegressor.optimismInputFeatureName', index=4,
      number=11, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='samplingScaleInputFeatureName', full_name='CoreML.Specification.BayesianProbitRegressor.samplingScaleInputFeatureName', index=5,
      number=12, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='samplingTruncationInputFeatureName', full_name='CoreML.Specification.BayesianProbitRegressor.samplingTruncationInputFeatureName', index=6,
      number=13, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='meanOutputFeatureName', full_name='CoreML.Specification.BayesianProbitRegressor.meanOutputFeatureName', index=7,
      number=20, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='varianceOutputFeatureName', full_name='CoreML.Specification.BayesianProbitRegressor.varianceOutputFeatureName', index=8,
      number=21, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='pessimisticProbabilityOutputFeatureName', full_name='CoreML.Specification.BayesianProbitRegressor.pessimisticProbabilityOutputFeatureName', index=9,
      number=22, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='sampledProbabilityOutputFeatureName', full_name='CoreML.Specification.BayesianProbitRegressor.sampledProbabilityOutputFeatureName', index=10,
      number=23, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_BAYESIANPROBITREGRESSOR_GAUSSIAN, _BAYESIANPROBITREGRESSOR_FEATUREVALUEWEIGHT, _BAYESIANPROBITREGRESSOR_FEATUREWEIGHT, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=56,
  serialized_end=856,
)

_BAYESIANPROBITREGRESSOR_GAUSSIAN.containing_type = _BAYESIANPROBITREGRESSOR
_BAYESIANPROBITREGRESSOR_FEATUREVALUEWEIGHT.fields_by_name['featureWeight'].message_type = _BAYESIANPROBITREGRESSOR_GAUSSIAN
_BAYESIANPROBITREGRESSOR_FEATUREVALUEWEIGHT.containing_type = _BAYESIANPROBITREGRESSOR
_BAYESIANPROBITREGRESSOR_FEATUREWEIGHT.fields_by_name['weights'].message_type = _BAYESIANPROBITREGRESSOR_FEATUREVALUEWEIGHT
_BAYESIANPROBITREGRESSOR_FEATUREWEIGHT.containing_type = _BAYESIANPROBITREGRESSOR
_BAYESIANPROBITREGRESSOR.fields_by_name['bias'].message_type = _BAYESIANPROBITREGRESSOR_GAUSSIAN
_BAYESIANPROBITREGRESSOR.fields_by_name['features'].message_type = _BAYESIANPROBITREGRESSOR_FEATUREWEIGHT
DESCRIPTOR.message_types_by_name['BayesianProbitRegressor'] = _BAYESIANPROBITREGRESSOR
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

BayesianProbitRegressor = _reflection.GeneratedProtocolMessageType('BayesianProbitRegressor', (_message.Message,), dict(

  Gaussian = _reflection.GeneratedProtocolMessageType('Gaussian', (_message.Message,), dict(
    DESCRIPTOR = _BAYESIANPROBITREGRESSOR_GAUSSIAN,
    __module__ = 'BayesianProbitRegressor_pb2'
    # @@protoc_insertion_point(class_scope:CoreML.Specification.BayesianProbitRegressor.Gaussian)
    ))
  ,

  FeatureValueWeight = _reflection.GeneratedProtocolMessageType('FeatureValueWeight', (_message.Message,), dict(
    DESCRIPTOR = _BAYESIANPROBITREGRESSOR_FEATUREVALUEWEIGHT,
    __module__ = 'BayesianProbitRegressor_pb2'
    # @@protoc_insertion_point(class_scope:CoreML.Specification.BayesianProbitRegressor.FeatureValueWeight)
    ))
  ,

  FeatureWeight = _reflection.GeneratedProtocolMessageType('FeatureWeight', (_message.Message,), dict(
    DESCRIPTOR = _BAYESIANPROBITREGRESSOR_FEATUREWEIGHT,
    __module__ = 'BayesianProbitRegressor_pb2'
    # @@protoc_insertion_point(class_scope:CoreML.Specification.BayesianProbitRegressor.FeatureWeight)
    ))
  ,
  DESCRIPTOR = _BAYESIANPROBITREGRESSOR,
  __module__ = 'BayesianProbitRegressor_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.BayesianProbitRegressor)
  ))
_sym_db.RegisterMessage(BayesianProbitRegressor)
_sym_db.RegisterMessage(BayesianProbitRegressor.Gaussian)
_sym_db.RegisterMessage(BayesianProbitRegressor.FeatureValueWeight)
_sym_db.RegisterMessage(BayesianProbitRegressor.FeatureWeight)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(descriptor_pb2.FileOptions(), _b('H\003'))
# @@protoc_insertion_point(module_scope)
