# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: Normalizer.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='Normalizer.proto',
  package='CoreML.Specification',
  syntax='proto3',
  serialized_pb=_b('\n\x10Normalizer.proto\x12\x14\x43oreML.Specification\"o\n\nNormalizer\x12;\n\x08normType\x18\x01 \x01(\x0e\x32).CoreML.Specification.Normalizer.NormType\"$\n\x08NormType\x12\x08\n\x04LMax\x10\x00\x12\x06\n\x02L1\x10\x01\x12\x06\n\x02L2\x10\x02\x42\x02H\x03\x62\x06proto3')
)



_NORMALIZER_NORMTYPE = _descriptor.EnumDescriptor(
  name='NormType',
  full_name='CoreML.Specification.Normalizer.NormType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='LMax', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='L1', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='L2', index=2, number=2,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=117,
  serialized_end=153,
)
_sym_db.RegisterEnumDescriptor(_NORMALIZER_NORMTYPE)


_NORMALIZER = _descriptor.Descriptor(
  name='Normalizer',
  full_name='CoreML.Specification.Normalizer',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='normType', full_name='CoreML.Specification.Normalizer.normType', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _NORMALIZER_NORMTYPE,
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=42,
  serialized_end=153,
)

_NORMALIZER.fields_by_name['normType'].enum_type = _NORMALIZER_NORMTYPE
_NORMALIZER_NORMTYPE.containing_type = _NORMALIZER
DESCRIPTOR.message_types_by_name['Normalizer'] = _NORMALIZER
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Normalizer = _reflection.GeneratedProtocolMessageType('Normalizer', (_message.Message,), dict(
  DESCRIPTOR = _NORMALIZER,
  __module__ = 'Normalizer_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.Normalizer)
  ))
_sym_db.RegisterMessage(Normalizer)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(descriptor_pb2.FileOptions(), _b('H\003'))
# @@protoc_insertion_point(module_scope)
