# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: SVM.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import DataStructures_pb2 as DataStructures__pb2
try:
  FeatureTypes__pb2 = DataStructures__pb2.FeatureTypes__pb2
except AttributeError:
  FeatureTypes__pb2 = DataStructures__pb2.FeatureTypes_pb2

from .DataStructures_pb2 import *

DESCRIPTOR = _descriptor.FileDescriptor(
  name='SVM.proto',
  package='CoreML.Specification',
  syntax='proto3',
  serialized_pb=_b('\n\tSVM.proto\x12\x14\x43oreML.Specification\x1a\x14\x44\x61taStructures.proto\"\x0e\n\x0cLinearKernel\"\x1a\n\tRBFKernel\x12\r\n\x05gamma\x18\x01 \x01(\x01\"6\n\nPolyKernel\x12\x0e\n\x06\x64\x65gree\x18\x01 \x01(\x05\x12\t\n\x01\x63\x18\x02 \x01(\x01\x12\r\n\x05gamma\x18\x03 \x01(\x01\")\n\rSigmoidKernel\x12\r\n\x05gamma\x18\x01 \x01(\x01\x12\t\n\x01\x63\x18\x02 \x01(\x01\"\xfa\x01\n\x06Kernel\x12:\n\x0clinearKernel\x18\x01 \x01(\x0b\x32\".CoreML.Specification.LinearKernelH\x00\x12\x34\n\trbfKernel\x18\x02 \x01(\x0b\x32\x1f.CoreML.Specification.RBFKernelH\x00\x12\x36\n\npolyKernel\x18\x03 \x01(\x0b\x32 .CoreML.Specification.PolyKernelH\x00\x12<\n\rsigmoidKernel\x18\x04 \x01(\x0b\x32#.CoreML.Specification.SigmoidKernelH\x00\x42\x08\n\x06kernel\"*\n\nSparseNode\x12\r\n\x05index\x18\x01 \x01(\x05\x12\r\n\x05value\x18\x02 \x01(\x01\"?\n\x0cSparseVector\x12/\n\x05nodes\x18\x01 \x03(\x0b\x32 .CoreML.Specification.SparseNode\"K\n\x14SparseSupportVectors\x12\x33\n\x07vectors\x18\x01 \x03(\x0b\x32\".CoreML.Specification.SparseVector\"\x1d\n\x0b\x44\x65nseVector\x12\x0e\n\x06values\x18\x01 \x03(\x01\"I\n\x13\x44\x65nseSupportVectors\x12\x32\n\x07vectors\x18\x01 \x03(\x0b\x32!.CoreML.Specification.DenseVector\"\x1d\n\x0c\x43oefficients\x12\r\n\x05\x61lpha\x18\x01 \x03(\x01\"\xb5\x02\n\x16SupportVectorRegressor\x12,\n\x06kernel\x18\x01 \x01(\x0b\x32\x1c.CoreML.Specification.Kernel\x12J\n\x14sparseSupportVectors\x18\x02 \x01(\x0b\x32*.CoreML.Specification.SparseSupportVectorsH\x00\x12H\n\x13\x64\x65nseSupportVectors\x18\x03 \x01(\x0b\x32).CoreML.Specification.DenseSupportVectorsH\x00\x12\x38\n\x0c\x63oefficients\x18\x04 \x01(\x0b\x32\".CoreML.Specification.Coefficients\x12\x0b\n\x03rho\x18\x05 \x01(\x01\x42\x10\n\x0esupportVectors\"\x8b\x04\n\x17SupportVectorClassifier\x12,\n\x06kernel\x18\x01 \x01(\x0b\x32\x1c.CoreML.Specification.Kernel\x12&\n\x1enumberOfSupportVectorsPerClass\x18\x02 \x03(\x05\x12J\n\x14sparseSupportVectors\x18\x03 \x01(\x0b\x32*.CoreML.Specification.SparseSupportVectorsH\x00\x12H\n\x13\x64\x65nseSupportVectors\x18\x04 \x01(\x0b\x32).CoreML.Specification.DenseSupportVectorsH\x00\x12\x38\n\x0c\x63oefficients\x18\x05 \x03(\x0b\x32\".CoreML.Specification.Coefficients\x12\x0b\n\x03rho\x18\x06 \x03(\x01\x12\r\n\x05probA\x18\x07 \x03(\x01\x12\r\n\x05probB\x18\x08 \x03(\x01\x12?\n\x11stringClassLabels\x18\x64 \x01(\x0b\x32\".CoreML.Specification.StringVectorH\x01\x12=\n\x10int64ClassLabels\x18\x65 \x01(\x0b\x32!.CoreML.Specification.Int64VectorH\x01\x42\x10\n\x0esupportVectorsB\r\n\x0b\x43lassLabelsB\x02H\x03P\x00\x62\x06proto3')
  ,
  dependencies=[DataStructures__pb2.DESCRIPTOR,],
  public_dependencies=[DataStructures__pb2.DESCRIPTOR,])




_LINEARKERNEL = _descriptor.Descriptor(
  name='LinearKernel',
  full_name='CoreML.Specification.LinearKernel',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=57,
  serialized_end=71,
)


_RBFKERNEL = _descriptor.Descriptor(
  name='RBFKernel',
  full_name='CoreML.Specification.RBFKernel',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='gamma', full_name='CoreML.Specification.RBFKernel.gamma', index=0,
      number=1, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=73,
  serialized_end=99,
)


_POLYKERNEL = _descriptor.Descriptor(
  name='PolyKernel',
  full_name='CoreML.Specification.PolyKernel',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='degree', full_name='CoreML.Specification.PolyKernel.degree', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='c', full_name='CoreML.Specification.PolyKernel.c', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='gamma', full_name='CoreML.Specification.PolyKernel.gamma', index=2,
      number=3, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=101,
  serialized_end=155,
)


_SIGMOIDKERNEL = _descriptor.Descriptor(
  name='SigmoidKernel',
  full_name='CoreML.Specification.SigmoidKernel',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='gamma', full_name='CoreML.Specification.SigmoidKernel.gamma', index=0,
      number=1, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='c', full_name='CoreML.Specification.SigmoidKernel.c', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=157,
  serialized_end=198,
)


_KERNEL = _descriptor.Descriptor(
  name='Kernel',
  full_name='CoreML.Specification.Kernel',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='linearKernel', full_name='CoreML.Specification.Kernel.linearKernel', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='rbfKernel', full_name='CoreML.Specification.Kernel.rbfKernel', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='polyKernel', full_name='CoreML.Specification.Kernel.polyKernel', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='sigmoidKernel', full_name='CoreML.Specification.Kernel.sigmoidKernel', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='kernel', full_name='CoreML.Specification.Kernel.kernel',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=201,
  serialized_end=451,
)


_SPARSENODE = _descriptor.Descriptor(
  name='SparseNode',
  full_name='CoreML.Specification.SparseNode',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='index', full_name='CoreML.Specification.SparseNode.index', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='value', full_name='CoreML.Specification.SparseNode.value', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=453,
  serialized_end=495,
)


_SPARSEVECTOR = _descriptor.Descriptor(
  name='SparseVector',
  full_name='CoreML.Specification.SparseVector',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='nodes', full_name='CoreML.Specification.SparseVector.nodes', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=497,
  serialized_end=560,
)


_SPARSESUPPORTVECTORS = _descriptor.Descriptor(
  name='SparseSupportVectors',
  full_name='CoreML.Specification.SparseSupportVectors',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='vectors', full_name='CoreML.Specification.SparseSupportVectors.vectors', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=562,
  serialized_end=637,
)


_DENSEVECTOR = _descriptor.Descriptor(
  name='DenseVector',
  full_name='CoreML.Specification.DenseVector',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='values', full_name='CoreML.Specification.DenseVector.values', index=0,
      number=1, type=1, cpp_type=5, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=639,
  serialized_end=668,
)


_DENSESUPPORTVECTORS = _descriptor.Descriptor(
  name='DenseSupportVectors',
  full_name='CoreML.Specification.DenseSupportVectors',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='vectors', full_name='CoreML.Specification.DenseSupportVectors.vectors', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=670,
  serialized_end=743,
)


_COEFFICIENTS = _descriptor.Descriptor(
  name='Coefficients',
  full_name='CoreML.Specification.Coefficients',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='alpha', full_name='CoreML.Specification.Coefficients.alpha', index=0,
      number=1, type=1, cpp_type=5, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=745,
  serialized_end=774,
)


_SUPPORTVECTORREGRESSOR = _descriptor.Descriptor(
  name='SupportVectorRegressor',
  full_name='CoreML.Specification.SupportVectorRegressor',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='kernel', full_name='CoreML.Specification.SupportVectorRegressor.kernel', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='sparseSupportVectors', full_name='CoreML.Specification.SupportVectorRegressor.sparseSupportVectors', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='denseSupportVectors', full_name='CoreML.Specification.SupportVectorRegressor.denseSupportVectors', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='coefficients', full_name='CoreML.Specification.SupportVectorRegressor.coefficients', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='rho', full_name='CoreML.Specification.SupportVectorRegressor.rho', index=4,
      number=5, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='supportVectors', full_name='CoreML.Specification.SupportVectorRegressor.supportVectors',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=777,
  serialized_end=1086,
)


_SUPPORTVECTORCLASSIFIER = _descriptor.Descriptor(
  name='SupportVectorClassifier',
  full_name='CoreML.Specification.SupportVectorClassifier',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='kernel', full_name='CoreML.Specification.SupportVectorClassifier.kernel', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='numberOfSupportVectorsPerClass', full_name='CoreML.Specification.SupportVectorClassifier.numberOfSupportVectorsPerClass', index=1,
      number=2, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='sparseSupportVectors', full_name='CoreML.Specification.SupportVectorClassifier.sparseSupportVectors', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='denseSupportVectors', full_name='CoreML.Specification.SupportVectorClassifier.denseSupportVectors', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='coefficients', full_name='CoreML.Specification.SupportVectorClassifier.coefficients', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='rho', full_name='CoreML.Specification.SupportVectorClassifier.rho', index=5,
      number=6, type=1, cpp_type=5, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='probA', full_name='CoreML.Specification.SupportVectorClassifier.probA', index=6,
      number=7, type=1, cpp_type=5, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='probB', full_name='CoreML.Specification.SupportVectorClassifier.probB', index=7,
      number=8, type=1, cpp_type=5, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='stringClassLabels', full_name='CoreML.Specification.SupportVectorClassifier.stringClassLabels', index=8,
      number=100, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='int64ClassLabels', full_name='CoreML.Specification.SupportVectorClassifier.int64ClassLabels', index=9,
      number=101, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='supportVectors', full_name='CoreML.Specification.SupportVectorClassifier.supportVectors',
      index=0, containing_type=None, fields=[]),
    _descriptor.OneofDescriptor(
      name='ClassLabels', full_name='CoreML.Specification.SupportVectorClassifier.ClassLabels',
      index=1, containing_type=None, fields=[]),
  ],
  serialized_start=1089,
  serialized_end=1612,
)

_KERNEL.fields_by_name['linearKernel'].message_type = _LINEARKERNEL
_KERNEL.fields_by_name['rbfKernel'].message_type = _RBFKERNEL
_KERNEL.fields_by_name['polyKernel'].message_type = _POLYKERNEL
_KERNEL.fields_by_name['sigmoidKernel'].message_type = _SIGMOIDKERNEL
_KERNEL.oneofs_by_name['kernel'].fields.append(
  _KERNEL.fields_by_name['linearKernel'])
_KERNEL.fields_by_name['linearKernel'].containing_oneof = _KERNEL.oneofs_by_name['kernel']
_KERNEL.oneofs_by_name['kernel'].fields.append(
  _KERNEL.fields_by_name['rbfKernel'])
_KERNEL.fields_by_name['rbfKernel'].containing_oneof = _KERNEL.oneofs_by_name['kernel']
_KERNEL.oneofs_by_name['kernel'].fields.append(
  _KERNEL.fields_by_name['polyKernel'])
_KERNEL.fields_by_name['polyKernel'].containing_oneof = _KERNEL.oneofs_by_name['kernel']
_KERNEL.oneofs_by_name['kernel'].fields.append(
  _KERNEL.fields_by_name['sigmoidKernel'])
_KERNEL.fields_by_name['sigmoidKernel'].containing_oneof = _KERNEL.oneofs_by_name['kernel']
_SPARSEVECTOR.fields_by_name['nodes'].message_type = _SPARSENODE
_SPARSESUPPORTVECTORS.fields_by_name['vectors'].message_type = _SPARSEVECTOR
_DENSESUPPORTVECTORS.fields_by_name['vectors'].message_type = _DENSEVECTOR
_SUPPORTVECTORREGRESSOR.fields_by_name['kernel'].message_type = _KERNEL
_SUPPORTVECTORREGRESSOR.fields_by_name['sparseSupportVectors'].message_type = _SPARSESUPPORTVECTORS
_SUPPORTVECTORREGRESSOR.fields_by_name['denseSupportVectors'].message_type = _DENSESUPPORTVECTORS
_SUPPORTVECTORREGRESSOR.fields_by_name['coefficients'].message_type = _COEFFICIENTS
_SUPPORTVECTORREGRESSOR.oneofs_by_name['supportVectors'].fields.append(
  _SUPPORTVECTORREGRESSOR.fields_by_name['sparseSupportVectors'])
_SUPPORTVECTORREGRESSOR.fields_by_name['sparseSupportVectors'].containing_oneof = _SUPPORTVECTORREGRESSOR.oneofs_by_name['supportVectors']
_SUPPORTVECTORREGRESSOR.oneofs_by_name['supportVectors'].fields.append(
  _SUPPORTVECTORREGRESSOR.fields_by_name['denseSupportVectors'])
_SUPPORTVECTORREGRESSOR.fields_by_name['denseSupportVectors'].containing_oneof = _SUPPORTVECTORREGRESSOR.oneofs_by_name['supportVectors']
_SUPPORTVECTORCLASSIFIER.fields_by_name['kernel'].message_type = _KERNEL
_SUPPORTVECTORCLASSIFIER.fields_by_name['sparseSupportVectors'].message_type = _SPARSESUPPORTVECTORS
_SUPPORTVECTORCLASSIFIER.fields_by_name['denseSupportVectors'].message_type = _DENSESUPPORTVECTORS
_SUPPORTVECTORCLASSIFIER.fields_by_name['coefficients'].message_type = _COEFFICIENTS
_SUPPORTVECTORCLASSIFIER.fields_by_name['stringClassLabels'].message_type = DataStructures__pb2._STRINGVECTOR
_SUPPORTVECTORCLASSIFIER.fields_by_name['int64ClassLabels'].message_type = DataStructures__pb2._INT64VECTOR
_SUPPORTVECTORCLASSIFIER.oneofs_by_name['supportVectors'].fields.append(
  _SUPPORTVECTORCLASSIFIER.fields_by_name['sparseSupportVectors'])
_SUPPORTVECTORCLASSIFIER.fields_by_name['sparseSupportVectors'].containing_oneof = _SUPPORTVECTORCLASSIFIER.oneofs_by_name['supportVectors']
_SUPPORTVECTORCLASSIFIER.oneofs_by_name['supportVectors'].fields.append(
  _SUPPORTVECTORCLASSIFIER.fields_by_name['denseSupportVectors'])
_SUPPORTVECTORCLASSIFIER.fields_by_name['denseSupportVectors'].containing_oneof = _SUPPORTVECTORCLASSIFIER.oneofs_by_name['supportVectors']
_SUPPORTVECTORCLASSIFIER.oneofs_by_name['ClassLabels'].fields.append(
  _SUPPORTVECTORCLASSIFIER.fields_by_name['stringClassLabels'])
_SUPPORTVECTORCLASSIFIER.fields_by_name['stringClassLabels'].containing_oneof = _SUPPORTVECTORCLASSIFIER.oneofs_by_name['ClassLabels']
_SUPPORTVECTORCLASSIFIER.oneofs_by_name['ClassLabels'].fields.append(
  _SUPPORTVECTORCLASSIFIER.fields_by_name['int64ClassLabels'])
_SUPPORTVECTORCLASSIFIER.fields_by_name['int64ClassLabels'].containing_oneof = _SUPPORTVECTORCLASSIFIER.oneofs_by_name['ClassLabels']
DESCRIPTOR.message_types_by_name['LinearKernel'] = _LINEARKERNEL
DESCRIPTOR.message_types_by_name['RBFKernel'] = _RBFKERNEL
DESCRIPTOR.message_types_by_name['PolyKernel'] = _POLYKERNEL
DESCRIPTOR.message_types_by_name['SigmoidKernel'] = _SIGMOIDKERNEL
DESCRIPTOR.message_types_by_name['Kernel'] = _KERNEL
DESCRIPTOR.message_types_by_name['SparseNode'] = _SPARSENODE
DESCRIPTOR.message_types_by_name['SparseVector'] = _SPARSEVECTOR
DESCRIPTOR.message_types_by_name['SparseSupportVectors'] = _SPARSESUPPORTVECTORS
DESCRIPTOR.message_types_by_name['DenseVector'] = _DENSEVECTOR
DESCRIPTOR.message_types_by_name['DenseSupportVectors'] = _DENSESUPPORTVECTORS
DESCRIPTOR.message_types_by_name['Coefficients'] = _COEFFICIENTS
DESCRIPTOR.message_types_by_name['SupportVectorRegressor'] = _SUPPORTVECTORREGRESSOR
DESCRIPTOR.message_types_by_name['SupportVectorClassifier'] = _SUPPORTVECTORCLASSIFIER
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

LinearKernel = _reflection.GeneratedProtocolMessageType('LinearKernel', (_message.Message,), dict(
  DESCRIPTOR = _LINEARKERNEL,
  __module__ = 'SVM_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.LinearKernel)
  ))
_sym_db.RegisterMessage(LinearKernel)

RBFKernel = _reflection.GeneratedProtocolMessageType('RBFKernel', (_message.Message,), dict(
  DESCRIPTOR = _RBFKERNEL,
  __module__ = 'SVM_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.RBFKernel)
  ))
_sym_db.RegisterMessage(RBFKernel)

PolyKernel = _reflection.GeneratedProtocolMessageType('PolyKernel', (_message.Message,), dict(
  DESCRIPTOR = _POLYKERNEL,
  __module__ = 'SVM_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.PolyKernel)
  ))
_sym_db.RegisterMessage(PolyKernel)

SigmoidKernel = _reflection.GeneratedProtocolMessageType('SigmoidKernel', (_message.Message,), dict(
  DESCRIPTOR = _SIGMOIDKERNEL,
  __module__ = 'SVM_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.SigmoidKernel)
  ))
_sym_db.RegisterMessage(SigmoidKernel)

Kernel = _reflection.GeneratedProtocolMessageType('Kernel', (_message.Message,), dict(
  DESCRIPTOR = _KERNEL,
  __module__ = 'SVM_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.Kernel)
  ))
_sym_db.RegisterMessage(Kernel)

SparseNode = _reflection.GeneratedProtocolMessageType('SparseNode', (_message.Message,), dict(
  DESCRIPTOR = _SPARSENODE,
  __module__ = 'SVM_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.SparseNode)
  ))
_sym_db.RegisterMessage(SparseNode)

SparseVector = _reflection.GeneratedProtocolMessageType('SparseVector', (_message.Message,), dict(
  DESCRIPTOR = _SPARSEVECTOR,
  __module__ = 'SVM_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.SparseVector)
  ))
_sym_db.RegisterMessage(SparseVector)

SparseSupportVectors = _reflection.GeneratedProtocolMessageType('SparseSupportVectors', (_message.Message,), dict(
  DESCRIPTOR = _SPARSESUPPORTVECTORS,
  __module__ = 'SVM_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.SparseSupportVectors)
  ))
_sym_db.RegisterMessage(SparseSupportVectors)

DenseVector = _reflection.GeneratedProtocolMessageType('DenseVector', (_message.Message,), dict(
  DESCRIPTOR = _DENSEVECTOR,
  __module__ = 'SVM_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.DenseVector)
  ))
_sym_db.RegisterMessage(DenseVector)

DenseSupportVectors = _reflection.GeneratedProtocolMessageType('DenseSupportVectors', (_message.Message,), dict(
  DESCRIPTOR = _DENSESUPPORTVECTORS,
  __module__ = 'SVM_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.DenseSupportVectors)
  ))
_sym_db.RegisterMessage(DenseSupportVectors)

Coefficients = _reflection.GeneratedProtocolMessageType('Coefficients', (_message.Message,), dict(
  DESCRIPTOR = _COEFFICIENTS,
  __module__ = 'SVM_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.Coefficients)
  ))
_sym_db.RegisterMessage(Coefficients)

SupportVectorRegressor = _reflection.GeneratedProtocolMessageType('SupportVectorRegressor', (_message.Message,), dict(
  DESCRIPTOR = _SUPPORTVECTORREGRESSOR,
  __module__ = 'SVM_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.SupportVectorRegressor)
  ))
_sym_db.RegisterMessage(SupportVectorRegressor)

SupportVectorClassifier = _reflection.GeneratedProtocolMessageType('SupportVectorClassifier', (_message.Message,), dict(
  DESCRIPTOR = _SUPPORTVECTORCLASSIFIER,
  __module__ = 'SVM_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.SupportVectorClassifier)
  ))
_sym_db.RegisterMessage(SupportVectorClassifier)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(descriptor_pb2.FileOptions(), _b('H\003'))
# @@protoc_insertion_point(module_scope)
