# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ItemSimilarityRecommender.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import DataStructures_pb2 as DataStructures__pb2
try:
  FeatureTypes__pb2 = DataStructures__pb2.FeatureTypes__pb2
except AttributeError:
  FeatureTypes__pb2 = DataStructures__pb2.FeatureTypes_pb2

from .DataStructures_pb2 import *

DESCRIPTOR = _descriptor.FileDescriptor(
  name='ItemSimilarityRecommender.proto',
  package='CoreML.Specification',
  syntax='proto3',
  serialized_pb=_b('\n\x1fItemSimilarityRecommender.proto\x12\x14\x43oreML.Specification\x1a\x14\x44\x61taStructures.proto\"\xb2\x05\n\x19ItemSimilarityRecommender\x12Z\n\x14itemItemSimilarities\x18\x01 \x03(\x0b\x32<.CoreML.Specification.ItemSimilarityRecommender.SimilarItems\x12\x39\n\ritemStringIds\x18\x02 \x01(\x0b\x32\".CoreML.Specification.StringVector\x12\x37\n\x0citemInt64Ids\x18\x03 \x01(\x0b\x32!.CoreML.Specification.Int64Vector\x12\x1c\n\x14itemInputFeatureName\x18\n \x01(\t\x12*\n\"numRecommendationsInputFeatureName\x18\x0b \x01(\t\x12\'\n\x1fitemRestrictionInputFeatureName\x18\x0c \x01(\t\x12%\n\x1ditemExclusionInputFeatureName\x18\r \x01(\t\x12,\n$recommendedItemListOutputFeatureName\x18\x14 \x01(\t\x12-\n%recommendedItemScoreOutputFeatureName\x18\x15 \x01(\t\x1a\x38\n\rConnectedItem\x12\x0e\n\x06itemId\x18\x01 \x01(\x04\x12\x17\n\x0fsimilarityScore\x18\x02 \x01(\x01\x1a\x93\x01\n\x0cSimilarItems\x12\x0e\n\x06itemId\x18\x01 \x01(\x04\x12V\n\x0fsimilarItemList\x18\x02 \x03(\x0b\x32=.CoreML.Specification.ItemSimilarityRecommender.ConnectedItem\x12\x1b\n\x13itemScoreAdjustment\x18\x03 \x01(\x01\x42\x02H\x03P\x00\x62\x06proto3')
  ,
  dependencies=[DataStructures__pb2.DESCRIPTOR,],
  public_dependencies=[DataStructures__pb2.DESCRIPTOR,])




_ITEMSIMILARITYRECOMMENDER_CONNECTEDITEM = _descriptor.Descriptor(
  name='ConnectedItem',
  full_name='CoreML.Specification.ItemSimilarityRecommender.ConnectedItem',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='itemId', full_name='CoreML.Specification.ItemSimilarityRecommender.ConnectedItem.itemId', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='similarityScore', full_name='CoreML.Specification.ItemSimilarityRecommender.ConnectedItem.similarityScore', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=564,
  serialized_end=620,
)

_ITEMSIMILARITYRECOMMENDER_SIMILARITEMS = _descriptor.Descriptor(
  name='SimilarItems',
  full_name='CoreML.Specification.ItemSimilarityRecommender.SimilarItems',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='itemId', full_name='CoreML.Specification.ItemSimilarityRecommender.SimilarItems.itemId', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='similarItemList', full_name='CoreML.Specification.ItemSimilarityRecommender.SimilarItems.similarItemList', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='itemScoreAdjustment', full_name='CoreML.Specification.ItemSimilarityRecommender.SimilarItems.itemScoreAdjustment', index=2,
      number=3, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=623,
  serialized_end=770,
)

_ITEMSIMILARITYRECOMMENDER = _descriptor.Descriptor(
  name='ItemSimilarityRecommender',
  full_name='CoreML.Specification.ItemSimilarityRecommender',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='itemItemSimilarities', full_name='CoreML.Specification.ItemSimilarityRecommender.itemItemSimilarities', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='itemStringIds', full_name='CoreML.Specification.ItemSimilarityRecommender.itemStringIds', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='itemInt64Ids', full_name='CoreML.Specification.ItemSimilarityRecommender.itemInt64Ids', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='itemInputFeatureName', full_name='CoreML.Specification.ItemSimilarityRecommender.itemInputFeatureName', index=3,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='numRecommendationsInputFeatureName', full_name='CoreML.Specification.ItemSimilarityRecommender.numRecommendationsInputFeatureName', index=4,
      number=11, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='itemRestrictionInputFeatureName', full_name='CoreML.Specification.ItemSimilarityRecommender.itemRestrictionInputFeatureName', index=5,
      number=12, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='itemExclusionInputFeatureName', full_name='CoreML.Specification.ItemSimilarityRecommender.itemExclusionInputFeatureName', index=6,
      number=13, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='recommendedItemListOutputFeatureName', full_name='CoreML.Specification.ItemSimilarityRecommender.recommendedItemListOutputFeatureName', index=7,
      number=20, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='recommendedItemScoreOutputFeatureName', full_name='CoreML.Specification.ItemSimilarityRecommender.recommendedItemScoreOutputFeatureName', index=8,
      number=21, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_ITEMSIMILARITYRECOMMENDER_CONNECTEDITEM, _ITEMSIMILARITYRECOMMENDER_SIMILARITEMS, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=80,
  serialized_end=770,
)

_ITEMSIMILARITYRECOMMENDER_CONNECTEDITEM.containing_type = _ITEMSIMILARITYRECOMMENDER
_ITEMSIMILARITYRECOMMENDER_SIMILARITEMS.fields_by_name['similarItemList'].message_type = _ITEMSIMILARITYRECOMMENDER_CONNECTEDITEM
_ITEMSIMILARITYRECOMMENDER_SIMILARITEMS.containing_type = _ITEMSIMILARITYRECOMMENDER
_ITEMSIMILARITYRECOMMENDER.fields_by_name['itemItemSimilarities'].message_type = _ITEMSIMILARITYRECOMMENDER_SIMILARITEMS
_ITEMSIMILARITYRECOMMENDER.fields_by_name['itemStringIds'].message_type = DataStructures__pb2._STRINGVECTOR
_ITEMSIMILARITYRECOMMENDER.fields_by_name['itemInt64Ids'].message_type = DataStructures__pb2._INT64VECTOR
DESCRIPTOR.message_types_by_name['ItemSimilarityRecommender'] = _ITEMSIMILARITYRECOMMENDER
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ItemSimilarityRecommender = _reflection.GeneratedProtocolMessageType('ItemSimilarityRecommender', (_message.Message,), dict(

  ConnectedItem = _reflection.GeneratedProtocolMessageType('ConnectedItem', (_message.Message,), dict(
    DESCRIPTOR = _ITEMSIMILARITYRECOMMENDER_CONNECTEDITEM,
    __module__ = 'ItemSimilarityRecommender_pb2'
    # @@protoc_insertion_point(class_scope:CoreML.Specification.ItemSimilarityRecommender.ConnectedItem)
    ))
  ,

  SimilarItems = _reflection.GeneratedProtocolMessageType('SimilarItems', (_message.Message,), dict(
    DESCRIPTOR = _ITEMSIMILARITYRECOMMENDER_SIMILARITEMS,
    __module__ = 'ItemSimilarityRecommender_pb2'
    # @@protoc_insertion_point(class_scope:CoreML.Specification.ItemSimilarityRecommender.SimilarItems)
    ))
  ,
  DESCRIPTOR = _ITEMSIMILARITYRECOMMENDER,
  __module__ = 'ItemSimilarityRecommender_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.ItemSimilarityRecommender)
  ))
_sym_db.RegisterMessage(ItemSimilarityRecommender)
_sym_db.RegisterMessage(ItemSimilarityRecommender.ConnectedItem)
_sym_db.RegisterMessage(ItemSimilarityRecommender.SimilarItems)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(descriptor_pb2.FileOptions(), _b('H\003'))
# @@protoc_insertion_point(module_scope)
