# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: NonMaximumSuppression.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import DataStructures_pb2 as DataStructures__pb2
try:
  FeatureTypes__pb2 = DataStructures__pb2.FeatureTypes__pb2
except AttributeError:
  FeatureTypes__pb2 = DataStructures__pb2.FeatureTypes_pb2

from .DataStructures_pb2 import *

DESCRIPTOR = _descriptor.FileDescriptor(
  name='NonMaximumSuppression.proto',
  package='CoreML.Specification',
  syntax='proto3',
  serialized_pb=_b('\n\x1bNonMaximumSuppression.proto\x12\x14\x43oreML.Specification\x1a\x14\x44\x61taStructures.proto\"\xc0\x04\n\x15NonMaximumSuppression\x12\x46\n\x07pickTop\x18\x01 \x01(\x0b\x32\x33.CoreML.Specification.NonMaximumSuppression.PickTopH\x00\x12?\n\x11stringClassLabels\x18\x64 \x01(\x0b\x32\".CoreML.Specification.StringVectorH\x01\x12=\n\x10int64ClassLabels\x18\x65 \x01(\x0b\x32!.CoreML.Specification.Int64VectorH\x01\x12\x14\n\x0ciouThreshold\x18n \x01(\x01\x12\x1b\n\x13\x63onfidenceThreshold\x18o \x01(\x01\x12#\n\x1a\x63onfidenceInputFeatureName\x18\xc8\x01 \x01(\t\x12$\n\x1b\x63oordinatesInputFeatureName\x18\xc9\x01 \x01(\t\x12%\n\x1ciouThresholdInputFeatureName\x18\xca\x01 \x01(\t\x12,\n#confidenceThresholdInputFeatureName\x18\xcb\x01 \x01(\t\x12$\n\x1b\x63onfidenceOutputFeatureName\x18\xd2\x01 \x01(\t\x12%\n\x1c\x63oordinatesOutputFeatureName\x18\xd3\x01 \x01(\t\x1a\x1b\n\x07PickTop\x12\x10\n\x08perClass\x18\x01 \x01(\x08\x42\x13\n\x11SuppressionMethodB\r\n\x0b\x43lassLabelsB\x02H\x03P\x00\x62\x06proto3')
  ,
  dependencies=[DataStructures__pb2.DESCRIPTOR,],
  public_dependencies=[DataStructures__pb2.DESCRIPTOR,])




_NONMAXIMUMSUPPRESSION_PICKTOP = _descriptor.Descriptor(
  name='PickTop',
  full_name='CoreML.Specification.NonMaximumSuppression.PickTop',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='perClass', full_name='CoreML.Specification.NonMaximumSuppression.PickTop.perClass', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=589,
  serialized_end=616,
)

_NONMAXIMUMSUPPRESSION = _descriptor.Descriptor(
  name='NonMaximumSuppression',
  full_name='CoreML.Specification.NonMaximumSuppression',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='pickTop', full_name='CoreML.Specification.NonMaximumSuppression.pickTop', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='stringClassLabels', full_name='CoreML.Specification.NonMaximumSuppression.stringClassLabels', index=1,
      number=100, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='int64ClassLabels', full_name='CoreML.Specification.NonMaximumSuppression.int64ClassLabels', index=2,
      number=101, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='iouThreshold', full_name='CoreML.Specification.NonMaximumSuppression.iouThreshold', index=3,
      number=110, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='confidenceThreshold', full_name='CoreML.Specification.NonMaximumSuppression.confidenceThreshold', index=4,
      number=111, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='confidenceInputFeatureName', full_name='CoreML.Specification.NonMaximumSuppression.confidenceInputFeatureName', index=5,
      number=200, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='coordinatesInputFeatureName', full_name='CoreML.Specification.NonMaximumSuppression.coordinatesInputFeatureName', index=6,
      number=201, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='iouThresholdInputFeatureName', full_name='CoreML.Specification.NonMaximumSuppression.iouThresholdInputFeatureName', index=7,
      number=202, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='confidenceThresholdInputFeatureName', full_name='CoreML.Specification.NonMaximumSuppression.confidenceThresholdInputFeatureName', index=8,
      number=203, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='confidenceOutputFeatureName', full_name='CoreML.Specification.NonMaximumSuppression.confidenceOutputFeatureName', index=9,
      number=210, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='coordinatesOutputFeatureName', full_name='CoreML.Specification.NonMaximumSuppression.coordinatesOutputFeatureName', index=10,
      number=211, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_NONMAXIMUMSUPPRESSION_PICKTOP, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='SuppressionMethod', full_name='CoreML.Specification.NonMaximumSuppression.SuppressionMethod',
      index=0, containing_type=None, fields=[]),
    _descriptor.OneofDescriptor(
      name='ClassLabels', full_name='CoreML.Specification.NonMaximumSuppression.ClassLabels',
      index=1, containing_type=None, fields=[]),
  ],
  serialized_start=76,
  serialized_end=652,
)

_NONMAXIMUMSUPPRESSION_PICKTOP.containing_type = _NONMAXIMUMSUPPRESSION
_NONMAXIMUMSUPPRESSION.fields_by_name['pickTop'].message_type = _NONMAXIMUMSUPPRESSION_PICKTOP
_NONMAXIMUMSUPPRESSION.fields_by_name['stringClassLabels'].message_type = DataStructures__pb2._STRINGVECTOR
_NONMAXIMUMSUPPRESSION.fields_by_name['int64ClassLabels'].message_type = DataStructures__pb2._INT64VECTOR
_NONMAXIMUMSUPPRESSION.oneofs_by_name['SuppressionMethod'].fields.append(
  _NONMAXIMUMSUPPRESSION.fields_by_name['pickTop'])
_NONMAXIMUMSUPPRESSION.fields_by_name['pickTop'].containing_oneof = _NONMAXIMUMSUPPRESSION.oneofs_by_name['SuppressionMethod']
_NONMAXIMUMSUPPRESSION.oneofs_by_name['ClassLabels'].fields.append(
  _NONMAXIMUMSUPPRESSION.fields_by_name['stringClassLabels'])
_NONMAXIMUMSUPPRESSION.fields_by_name['stringClassLabels'].containing_oneof = _NONMAXIMUMSUPPRESSION.oneofs_by_name['ClassLabels']
_NONMAXIMUMSUPPRESSION.oneofs_by_name['ClassLabels'].fields.append(
  _NONMAXIMUMSUPPRESSION.fields_by_name['int64ClassLabels'])
_NONMAXIMUMSUPPRESSION.fields_by_name['int64ClassLabels'].containing_oneof = _NONMAXIMUMSUPPRESSION.oneofs_by_name['ClassLabels']
DESCRIPTOR.message_types_by_name['NonMaximumSuppression'] = _NONMAXIMUMSUPPRESSION
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

NonMaximumSuppression = _reflection.GeneratedProtocolMessageType('NonMaximumSuppression', (_message.Message,), dict(

  PickTop = _reflection.GeneratedProtocolMessageType('PickTop', (_message.Message,), dict(
    DESCRIPTOR = _NONMAXIMUMSUPPRESSION_PICKTOP,
    __module__ = 'NonMaximumSuppression_pb2'
    # @@protoc_insertion_point(class_scope:CoreML.Specification.NonMaximumSuppression.PickTop)
    ))
  ,
  DESCRIPTOR = _NONMAXIMUMSUPPRESSION,
  __module__ = 'NonMaximumSuppression_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.NonMaximumSuppression)
  ))
_sym_db.RegisterMessage(NonMaximumSuppression)
_sym_db.RegisterMessage(NonMaximumSuppression.PickTop)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(descriptor_pb2.FileOptions(), _b('H\003'))
# @@protoc_insertion_point(module_scope)
